/*
  Name: linearplot.c
  Copyright: Copyright  Robert A Bradley 10/9/2005
  Author: Robert A Bradley
  Date: 21/09/2005 14:30
  Description: Point plotter with linear fit and error calculation
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "pointlib.h"
#include "pcpfile.h"

#define USE_GNUPLOT
#ifdef USE_GNUPLOT
#include "gnuplot_i.c"
#include "gnuplot_ext.h"
#endif
#define square(x) ((x)*(x))

#define showLaTeX 0

int main(int argc, char *argv[])
{
    int i=0;
    char filename[1024];
    if (argc<2) {
        printf("Please enter a filename: ");
        scanf("%1023s",filename);
    }
    else {
        strncpy(filename,argv[1],1024);
    }
    PCPFile *pcp=InitPCPFile();
    if (!pcp) {
        return EXIT_FAILURE;
    }
    if (LoadFile(filename,pcp)) {
        DisposePCPFile(pcp);
        printf("Error loading PCP file\n");
        return EXIT_FAILURE;
    }
    Point *points=pcp->points;
    int pCount=pcp->count;
    printf("Loaded file '%s' sucessfully, %d points present\n",filename,pCount);
    double dblCount=pCount;
//-----
    double sumX=0;
    double sumY=0;
    double sumX2=0;
    double sumXY=0;

    for (i=0;i<pCount;i++) {
        sumX+=points[i].x;
        sumY+=points[i].y;
        sumX2+=square(points[i].x);
        sumXY+=points[i].x*points[i].y;
    }
    double gradient=sumXY-sumX*sumY/dblCount;
    gradient/=sumX2-(sumX*sumX/dblCount);
    double c=(sumY-gradient*sumX)/dblCount;
    
    double d=0,d2=0,varY=0,varX=0; // d=standard deviation of y-yfit, d2=variance
    double ax=0,ay=0,ayf=0,ax2=0; // ax, ay, ayf, ax2 - mean values of x, y, yfit, x^2
    double sumDY_YF2=0; // Sum of (y[i]-yfit[i])^2 over i
    for (i=0;i<pCount;i++) {
        points[i].fit_y=gradient*points[i].x+c;
        ax+=points[i].x;
        ay+=points[i].y;
        ayf+=points[i].fit_y;
        ax2+=square(points[i].x);
        sumDY_YF2+=square(points[i].y-points[i].fit_y);
    }
    ax/=dblCount;
    double sum_dX2=ax2;
    ax2/=dblCount;
    ay/=dblCount;
    ayf/=dblCount;

    for (i=0;i<pCount;i++) {
        varY+=square(points[i].y-ay);
        varX+=square(points[i].x-ax);
    }
    sum_dX2-=dblCount*ax*ax;
    d2=sumDY_YF2/dblCount;
    double varGradient=sumDY_YF2/((dblCount-2)*sum_dX2);
    double varIntY=varGradient*ax2;
    varY=varY/dblCount;// Sample variance (Y)
    varX=varX/dblCount;// Sample variance (X)
    //double sd=sqrt(varY),sdX=sqrt(varX);
    double sdGradient=sqrt(varGradient);
    double sdIntY=sqrt(varIntY);
    d=sqrt(d2);
    double r=sqrt(1-d2/varY); // Correlation coefficient
    printf("Linear fit\n");
    printf("The standard deviation of fit=%e\n",d);
    printf("Correlation coefficient=%e\n\n",r);
    
    printf("xbar\t%e\n",ax);
    printf("ybar\t%e\n",ay);
    printf("Slope\t%e\t%e\n",gradient,sdGradient);
    printf("y intercept\t%e\t%e\n",c,sdIntY);
    printf("x intercept\t%e\t%e\n",(-c/gradient),sqrt(varIntY+(c*c/(gradient*gradient)*varGradient))/gradient);
    
    double rms=0; // RMS error in y (i.e. y-yfit)
    for (i=0;i<pCount;i++) {
        rms+=square(points[i].y-points[i].fit_y);
    }
    rms=sqrt(d/dblCount);
    // Note: "Err/sd" calculation is wrong
    printf("\nPoint\tx\ty\tfit\tErr/sd\n");
    for (i=0;i<pCount;i++) {
        double seYF2=varY+square(points[i].x)*varGradient+varX*square(gradient)+varIntY;
        double se=sqrt(seYF2);// Standard error in y-yfit for the current point
        printf("%d\t%1.6e\t%1.6e\t%1.6e\t%1.6e\n",i+1,points[i].x,points[i].y,points[i].fit_y,(points[i].y-points[i].fit_y)/se);
    }
    
#ifdef USE_GNUPLOT
    // Plot points and best-fit line
    gnuplot_ctrl *gp;
    if (!(gp=gnuplot_init())) {
        //failed to initialise...
        gnuplot_close(gp);
        printf("GNUPlot failed to start.  Plotting failed.\n");
    }
    else {
        gnuplot_setstyle(gp,"points");
        gnuplot_set_xlabel(gp,"x");
        gnuplot_set_ylabel(gp,"y");
        gnuplot_plot_Point_xy(gp,points,pCount,"Points");
        char cmd[4096];
        sprintf(cmd,"replot %f*x+%f",gradient,c);
        strncat(cmd," title 'Linear fit' with lines\n",4096-strlen(cmd));
        gnuplot_cmd(gp,cmd);
        char q=0; // Temporary variable used to check for X key
        printf("Press space (and enter) to continue...\n");
        while (q!=' ')
            scanf("%c",&q);
        gnuplot_close(gp);
    }
#else
    printf("gnuplot_i support not enabled for this operating system.\n");
#endif
    // Cleanup resources and exit
    DisposePCPFile(pcp);
    return EXIT_SUCCESS;
}
