/*
  Name: pcpfile.h
  Copyright: Copyright  Robert A Bradley 9/9/2005
  Author: Robert A Bradley
  Date: 21/09/2005 14:28
  Description: PCP file handling routines
*/

#ifndef _PCPFILE_H
#define _PCPFILE_H
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "pointlib.h"

typedef struct PCPFile {
        int count;
        int plotType;
        int weightType;
        Point *points;
        } PCPFile;
        
/*
PCPFile* InitPCPFile()

Creates an empty PCPFile structure, with the "points" pointer set to NULL.
Returns NULL if malloc fails.
*/

PCPFile* InitPCPFile() {
        PCPFile *pcp=(PCPFile*) malloc(sizeof(PCPFile));
        if (!pcp) {
            return NULL;
        }
        pcp->count=0;
        pcp->points=NULL;
        pcp->plotType=0;
        pcp->weightType=0;
        return pcp;
}

/*
char* GetLine(char *line,int len,FILE *fin)

Like fgets, but ignores any extra characters before the new line.
*/

char* GetLine(char *line,int len,FILE *fin) {
    char *ret;int i=0;
    if (!(ret=fgets(line,len,fin))) {
        return NULL;
    }
    if (strlen(line)>0 && line[strlen(line)-1]!='\n') {
        char tmp='\0';
        while (tmp!='\n' && i!=EOF) {
            i=fscanf(fin,"%c",&tmp);
        }
    }
    line[strlen(line)-1]=0;
    line[len-1]=0;
    return ret;
}

/*
int CountChars(char *line,char chr)

Counts the number of occurrences of the character chr.
*/

int CountChars(char *line,char chr) {
    char *x=line;int i=0;
    while ((x=strchr(x,chr))) {
    x++;i++;
    }
    return i;
}

/*
int LoadCSVFile(char *filename,PCPFile *pcp)

Loads the contents of CSV file filename into pcp structure, allocating points.
Returns non-zero on failure:
        
        1        File open error
        2        malloc failed while allocating points array
*/

int LoadCSVFile(char *filename,PCPFile *pcp) {
    FILE *fin;
    long nline=1;
    char line[1024];
    if (!(fin=fopen(filename,"r"))) {
        printf("Can't open %s for reading!\n",filename);
        return 1;
    }
    Pointlist *points;
    if (!(points=MakePointlist())) {
        printf("Could not create Pointlist!\n");
        return 1;
    }
    GetLine(line,1024,fin);nline++;
    Point p;
    while (!feof(fin)) {
        char *b=GetLine(line,1024,fin);
        if (b) {
            int i=CountChars(b,',');
#if DEBUG
            printf("Line %ld, i=%d\n",nline,i);
#endif
            if (i>1) {
                sscanf(b,"%le,%le,%le",&p.x,&p.y,&p.err_y);
            }
            else if (i>0) {
                sscanf(b,"%le,%le",&p.x,&p.y);
                p.err_y=0;
            }
            else {
                printf("Bad input line: %ld\n",nline);
            }
#if DEBUG
            printf("Scanf successful\n");
#endif
            AddPoint(points,p);
            line[0]=0;
        }
        nline++;
    }
    fclose(fin);
    pcp->points=PointlistToArray(points);
    pcp->count=GetPointCount(points);
    FreePointlist(points);
    if (!(pcp->points))
        return 2;
    return 0;
}

/*
int LoadPCPFile(char *filename,PCPFile *pcp)

Loads the contents of file referenced by filename into pcp, allocating points.
Returns non-zero on failure:
        
        1        File open error
        2        malloc failed while allocating points array
*/

int LoadPCPFile(char *filename,PCPFile *pcp) {
    FILE *fin;
    int fscanret=0;
    if (!(fin=fopen(filename,"r"))) {
        printf("Can't open %s for reading!\n",filename);
        return 1;
    }
    int i;
    fscanf(fin,"%d%d%d\n",&(pcp->count),&(pcp->plotType),&(pcp->weightType));
    Point *points=(Point*) malloc(sizeof(Point)*pcp->count);
    pcp->points=points;
    if (!points) {
       return 2;
    }
    for (i=0;i<pcp->count;i++) {
        Point *p=&points[i];
        if (pcp->weightType==0) {
            fscanret=fscanf(fin,"%lg%lg\n",&(p->x),&(p->y));
            p->err_y=0;
        }
        else {
            fscanret=fscanf(fin,"%lg%lg%lg\n",&(*p).x,&(*p).y,&(*p).err_y);
        }
        if (fscanret==EOF) {
            printf("Warning: file may be corrupted (%d of %d points read).",i+1,pcp->count);
            pcp->count=i+1;
            break;
        }
    }
    fclose(fin);
    return 0;
}

/*
int LoadFile(char *filename,PCPFile *pcp)

Wrapper for LoadPCPFile and LoadCSVFile - calls the appropriate function for the
file, depending on the file extension.
*/

int LoadFile(char *filename,PCPFile *pcp) {
    int end=strlen(filename);
    char ext[6];
    strcpy(ext,filename+end-4);
    if (!strcmp(ext,".csv")) {
        printf("Loading CSV file %s...\n",filename);
        return LoadCSVFile(filename,pcp);
    }
    printf("Loading PCP file %s...\n",filename);
    return LoadPCPFile(filename,pcp);
}

/*
int SavePCPFile(char *filename,PCPFile *pcp)

Save the contents of the pcp structure in PCP format to the file referenced by filename.
Returns non-zero on failure:
        
        1        File open error
*/

int SavePCPFile(char *filename,PCPFile *pcp) {
    FILE *fout;
    if (!(fout=fopen(filename,"w"))) {
        printf("Can't open %s for writing!\n",filename);
        return 1;
    }
    fprintf(fout,"%d %d %d\n",pcp->count,pcp->plotType,pcp->weightType);
    Point *points=pcp->points;
    int i;
    if (pcp->weightType==0) {
        for (i=0;i<pcp->count;i++) {
            fprintf(fout,"%g %g\n",points[i].x,points[i].y);
        }
    }
    else {
        for (i=0;i<pcp->count;i++) {
            fprintf(fout,"%g %g %g\n",points[i].x,points[i].y,points[i].err_y);
        }
    }
    fclose(fout);
    return 0;
}

/*
int SavePCPFileFromPointlist(char *filename,PCPFile *pcp,Pointlist *pl)

Save the contents of pl in PCP format to the file referenced by filename.  The pcp
pointer is only used to specify options such as plot type - the number of points and
their values are all obtained from the Pointlist.

This function is a alternative to doing a Pointlist conversion (saving up to 50%
of memory usage):
     
     pcp->points=PointlistToArray(pl);
     pcp->count=GetPointCount(pl);
     SavePCPFile("data.pcp",pcp);

Returns non-zero on failure:
        
        1        File open error
*/

int SavePCPFileFromPointlist(char *filename,PCPFile *pcp,Pointlist *pl) {
    FILE *fout;
    if (!(fout=fopen(filename,"w"))) {
        printf("Can't open %s for writing!\n",filename);
        return 1;
    }
    pcp->count=GetPointCount(pl);
    fprintf(fout,"%d %d %d\n",pcp->count,pcp->plotType,pcp->weightType);
    int i;
    if (pcp->weightType==0) {
        for (i=0;i<pcp->count;i++) {
            Point ps=GetPoint(pl,i);
            fprintf(fout,"%g %g\n",ps.x,ps.y);
        }
    }
    else {
        for (i=0;i<pcp->count;i++) {
            Point ps=GetPoint(pl,i);
            fprintf(fout,"%g %g %g\n",ps.x,ps.y,ps.err_y);
        }
    }
    fclose(fout);
    return 0;
}

/*
void DisposeStaticPCPFile(PCPFile *pcp)

Deallocates a static PCPFile structure, not including pcp.  Only use this if you
created your pcp object using:

       PCPFile pcp;

The InitPCPFile function uses malloc, so either use DisposePCPFile instead, or do
an explicit free(pcp) afterwards.
*/

void DisposeStaticPCPFile(PCPFile *pcp) {
     if (pcp) {
         if (pcp->points) {
             free(pcp->points);
             pcp->points=NULL;
         }
     }
}

/*
void DisposePCPFile(PCPFile *pcp)

Deallocates a PCPFile structure, *including* the pcp object.
Be careful not to double-free the points elsewhere in your code, e.g.

Point *points=blah();
pcp->points=points;
...
// To avoid double free, uncomment next line:
// points=NULL;
free(points);
DisposePCPFile(pcp); // Problem here!
*/

void DisposePCPFile(PCPFile *pcp) {
         if (pcp) {
             DisposeStaticPCPFile(pcp);
         free(pcp);
         pcp=NULL;
     }
}
#endif
