/*
  Name: polyplot.c
  Copyright: Copyright  Robert A Bradley 10/9/2005
  Author: Robert A Bradley
  Date: 21/09/2005 14:28
  Description: Point plotter with polynomial fit and error calculation
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "pointlib.h"
#include "pcpfile.h"

#define USE_GNUPLOT
#ifdef USE_GNUPLOT
#include "gnuplot_i.c"
#include "gnuplot_ext.h"
#endif

#define showLaTeX 0

double sum(Point *points,int dSize,int powX) {
    int j;
    double s=0;
    for (j=0;j<dSize;j++) {
        s+=(points[j].y*pow(points[j].x,powX));
    }
    return s;
}

double sumX(Point *points,int dSize,int powX) {
    int j;
    double s=0;
    for (j=0;j<dSize;j++) {
        s+=pow(points[j].x,powX);
    }
    return s;
}

double p(double *c,int coeff,double x) {
    int i;
    double y=0;
    for (i=0;i<=coeff;i++) {
        y+=pow(x,i)*c[i];
    }
    return y;
}

int GaussJordan(double *a, double *b, double *x, int coefficients) {
    double temp;
    int i,j,counter;
    double mat[coefficients][coefficients+1];
    
    /* Merge the matrix A and the vector B into one matrix */
    for (i=0; i<coefficients; i++) {
        for (j=0; j<coefficients;j++)
        mat[i][j] = *(a + i*coefficients + j);
        mat[i][coefficients] = *(b+i);
    }
    
    for (i=0;i<=coefficients-1;i++) {
        temp=mat[i][i];
        for (j=0;j<=coefficients;j++)
            mat[i][j]=mat[i][j]/temp;
        for (counter=i+1;counter<=coefficients-1;counter++) {
            temp=mat[counter][i];
            for (j=0;j<=coefficients;j++)
            mat[counter][j]=-1*temp*mat[i][j]+mat[counter][j];
        }
    }
    
    for (i=coefficients-1;i>=0;i--)
        for (counter=i-1;counter>=0;counter--)
        {
            temp=mat[counter][i];
            for (j=0;j<=coefficients;j++)
                mat[counter][j]=-1*temp*mat[i][j]+mat[counter][j];
        }
    
    for (i=0; i<coefficients; i++)
        x[i] = mat[i][coefficients];
    
    return 0;
}


int main(int argc, char *argv[])
{
    int i=0;
    char q=0; // Temporary variable used to check for X key
    char filename[1024];
    if (argc<2) {
        printf("Please enter a filename: ");
        scanf("%1023s",filename);
    }
    else {
        strncpy(filename,argv[1],1024);
    }
    PCPFile *pcp=InitPCPFile();
    if (!pcp) {
        return EXIT_FAILURE;
    }
    if ((i=LoadFile(filename,pcp))) {
        DisposePCPFile(pcp);
        printf("Error loading PCP file '%s'\n",filename);
        return EXIT_FAILURE;
    }
    Point *points=pcp->points; // Do not free points var - use DisposePCPFile
    int pCount=pcp->count;
    printf("Loaded file '%s' sucessfully, %d points present\n",filename,pCount);
    
//-----
    int k;
    int coefficients=0; // Number of coefficients
    while (coefficients>9 || coefficients<2) {
        printf("Please enter number of coefficients for fit: ");
        scanf("%d",&coefficients);
    }
    int cSize=coefficients+1;
    double a[cSize][cSize];
    double v[cSize],c[cSize];
    
    for (k=0;k<=coefficients;k++) {
        for (i=0;i<=coefficients;i++) {
            a[k][i]=sumX(points,pCount,i+k);
        }
        v[k]=sum(points,pCount,k);
    }
    
    // Solve the equations
    
    if (GaussJordan((double *) a, v, c, cSize)) {
        printf("Could not solve polynomial fit!\n");
        exit(1);
    }
    // Coefficients now in c[]...
#if showLaTeX
        char tmp[20],pm[4];
        double mant;
        int e;
        printf("LaTeX output:\n$");

        for (i=coefficients;i>=0;i--) {
            sprintf(tmp,"%e",c[i]);
            strcpy(pm,(c[i]<0) ? "":"+");
            char *pos=strchr(tmp,'e');
            sscanf(tmp,"%lf",&mant);
            sscanf(pos+1,"%d",&e);
            printf("%s%f*10^{%d}*x^{%d}",pm,mant,e,i);
        }
        printf("$\n");
#endif
    
    for (i=0;i<pCount;i++) {
        points[i].fit_y=p(c,coefficients,points[i].x);
    }
    
    double d=0;
    for (i=0;i<pCount;i++) {
        d+=pow(points[i].y-points[i].fit_y,2);
    }
    d=sqrt(d/pCount);
    printf("\nFit for polynomial of order %d\n",coefficients);
    printf("The rms error of this fit is%e\n\n",d);
    
    //printf("Point\tx\ty\tfit\ty-fit\t(y-fit)/rms\n");    
    printf(" Point       x              y              fit             y-fit          (y-fit)/rms\n");
    for (i=0;i<pCount;i++) {
        //printf("%d&%.5e&%.5e&%.5e&%.3e&%.3e\\\\\n",i+1,x[i],y[i],yf[i],y[i]-yf[i],(y[i]-yf[i])/d);
        printf("%4d %15.5e %15.5e %15.5e %15.3e %15.3e\n",i+1,points[i].x,points[i].y,points[i].fit_y,points[i].y-points[i].fit_y,(points[i].y-points[i].fit_y)/d);
    }
    printf("\nPolynomial Coefficients:\n");
    for (i=0;i<=coefficients;i++) {
        printf("%4d %15.6e\n",i,c[i]);
    }
    printf("\n");
    
#ifdef USE_GNUPLOT
    gnuplot_ctrl *gp;
    if (!(gp=gnuplot_init())) {
        //failed to initialise...
        gnuplot_close(gp);
        printf("GNUPlot failed to start.  Plotting failed.\n");
    }
    else {
        gnuplot_setstyle(gp,"points");
        gnuplot_set_xlabel(gp,"x");
        gnuplot_set_ylabel(gp,"y");
        gnuplot_plot_Point_xy(gp,points,pCount,"Points");
        char cmd[4096]="replot ",str[4000],strtmp[30];
        int len=4000;
        for (i=0;i<=coefficients;i++) {
            int n=sprintf(strtmp,"+(%g)*(x**%d)",c[i],i);
            strncat(str,strtmp,len);
            len-=n;
        }
        strncat(cmd,str,4096-strlen(cmd));
        strncat(cmd," title 'Polynomial fit' with lines\n",4096-strlen(cmd));
        gnuplot_cmd(gp,cmd);
        printf("Press space (and enter) to continue...\n");
        while (q!=' ')
            scanf("%c",&q);
        gnuplot_close(gp);
    }
#else
    printf("gnuplot_i support not enabled for this operating system.\n");
#endif
    DisposePCPFile(pcp);
    return EXIT_SUCCESS;
}
